@@ -15,11 +15,11 @@
 ;; You should have received a copy of the GNU General Public License
 ;; along with this program. If not, see <http://www.gnu.org/licenses/>.
 
-;; Author:           Martin Edström <meedstrom91@gmail.com>
+;; Author:           Martin Edström <meedstrom@runbox.eu>
 ;; URL:              https://github.com/meedstrom/el-job
 ;; Created:          2024-10-30
 ;; Keywords:         processes
-;; Package-Version:  2.3.1
+;; Package-Version:  2.4.0
 ;; Package-Requires: ((emacs "29.1"))
 
 ;;; Commentary:
@@ -48,12 +48,16 @@
 ;;; Code:
 
 ;; Wish-list:
-;; - Figure out how to use `dump-emacs-portable' after loading any number of
-;;   features so we can reuse that to start the subprocesses instantly.  Could
-;;   even simplify to a sentinel-based workflow then, no polling.
+;; - Figure out how to allow more than one list of INPUTS.  That'd give this
+;;   library much wider applicability!
 
-;; - Figure out how to allow more than one list of INPUTS.  That'd let this
-;;   library be applicable to many more situations.
+;; - Figure out how to use `dump-emacs-portable' after loading all desired
+;;   features, so we can reuse that to start the subprocesses nigh-instantly.
+;;   Could even simplify to a sentinel-based workflow then, no polling.
+
+;; - Improve the error message in some cases when it is unhelpful and only says
+;;   something like "Error running timer: listp Error:".  I surmise the process
+;;   buffer contents are not in a Lisp-readable form then.
 
 (require 'cl-lib)
 (require 'el-job-child)
@@ -263,7 +267,8 @@ being saddled with a huge item in addition to the average workload."
                     (setq this-sublist nil)
                     (push item items)))))))
         (if (length= sublists 0)
-            (progn ;; Degrade gracefully
+            (progn
+              ;; Degrade gracefully (but I think we never arrive here)
               (fset 'el-job--split-optimally #'el-job--split-evenly)
               (cl-assert (not (length= sublists 0))))
           ;; Spread leftovers equally
@@ -373,11 +378,13 @@ process that locates files \(by inspecting `load-history', via
 
 INPUTS is a list that will be split by up to the output of
 `num-processors', and this determines how many subprocesses will spawn.
-If INPUTS is omitted, only one subprocess will spawn.
 
 INPUTS can also be a function that returns a list.  In this case, the
 function is deferred until needed, possibly saving on compute.
 
+If INPUTS returns nil, do nothing and return the symbol
+`inputs-were-empty'.
+
 
 The subprocesses have no access to current Emacs state.  The only way
 they can affect current state, is if FUNCALL-PER-INPUT returns data,
@@ -446,29 +453,31 @@ For debugging, see these commands:
               ('wait (setf .queued-inputs (append inputs .queued-inputs))))
           (setf .queued-inputs inputs)
           (setq do-exec t))
-        (when do-exec
-          (setf .callback callback)
-          ;; Prevent spawning a dozen processes when you'll use only one or two
-          (let ((machine-cores (max 1 (1- (num-processors)))))
-            (setf .n-cores-to-use (if (length< .queued-inputs machine-cores)
-                                      (length .queued-inputs)
-                                    machine-cores))
-            (when (or (length< .ready .n-cores-to-use)
-                      (not (cl-every #'process-live-p .ready)))
-              (setq do-respawn t)))
-          (let ((new-spawn-args (list job
-                                      load-features
-                                      inject-vars
-                                      funcall-per-input)))
-            (unless (= (sxhash (cdr .spawn-args))
-                       (sxhash (cdr new-spawn-args)))
-              (setf .spawn-args new-spawn-args)
-              (el-job--dbg 2 "New arguments, resetting processes for %s" id)
-              (setq do-respawn t)))
-          (when do-respawn
-            (el-job--disable job)
-            (apply #'el-job--spawn-processes .spawn-args))
-          (el-job--exec-workload job))))))
+        (if (null .queued-inputs)
+            'inputs-were-empty
+          (when do-exec
+            (setf .callback callback)
+            ;; Prevent spawning a dozen processes when we need only one or two
+            (let ((machine-cores (max 1 (1- (num-processors)))))
+              (setf .n-cores-to-use (if (length< .queued-inputs machine-cores)
+                                        (length .queued-inputs)
+                                      machine-cores))
+              (when (or (length< .ready .n-cores-to-use)
+                        (not (cl-every #'process-live-p .ready)))
+                (setq do-respawn t)))
+            (let ((new-spawn-args (list job
+                                        load-features
+                                        inject-vars
+                                        funcall-per-input)))
+              (unless (= (sxhash (cdr .spawn-args))
+                         (sxhash (cdr new-spawn-args)))
+                (setf .spawn-args new-spawn-args)
+                (el-job--dbg 2 "New arguments, resetting processes for %s" id)
+                (setq do-respawn t)))
+            (when do-respawn
+              (el-job--disable job)
+              (apply #'el-job--spawn-processes .spawn-args))
+            (el-job--exec-workload job)))))))
 
 (defvar-local el-job-here nil)
 (defun el-job--spawn-processes (job load-features inject-vars funcall-per-input)
@@ -582,10 +591,10 @@ should trigger `el-job--handle-output'."
 
 ;; So, poll.  We do a chain of timers that successively ups the delay.
 ;; To see what the delays would be, eval:
-;; (--map (/ it (float (ash 1 5))) (-iterate '1+ 1 42))
+;; (--map (/ it 32.0) (-iterate '1+ 1 42))
 
 ;; And to see the cumulative sums:
-;; (-reductions '+ (--map (/ it (float (ash 1 5))) (-iterate '1+ 1 42)))
+;; (-reductions '+ (--map (/ it 32.0) (-iterate '1+ 1 42)))
 
 ;; As you can see, we do 7 polls inside the first second,
 ;; but spread out the last 7 polls between T=20s and T=30s.
@@ -609,7 +618,7 @@ after a short delay.  N is the count of checks done so far."
           (if (<= n 42)
               (setf (el-job-timer el-job-here)
                     (run-with-timer
-                     (/ n (float (ash 1 5))) nil #'el-job--poll (1+ n) busy-bufs))
+                     (/ n 32.0) nil #'el-job--poll (1+ n) busy-bufs))
             (el-job--disable el-job-here)
             (message "el-job: Timed out, was busy for 30+ seconds: %s"
                      (el-job-id el-job-here)))
@@ -624,6 +633,7 @@ after a short delay.  N is the count of checks done so far."
       (el-job--disable job)
       (el-job--dbg 2 "Reaped idle processes for %s" (el-job-id job)))))
 
+;; TODO: Consider using `with-local-quit', because this fires on a timer
 (defun el-job--handle-output ()
   "Handle output in current buffer.
 
@@ -773,20 +783,20 @@ Safely return nil otherwise, whether or not ID is known."
 
 ;;; Ok, don't need to break convention anymore.
 
-(define-obsolete-function-alias 'el-job:id             #'el-job-id "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:callback       #'el-job-callback "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:id             #'el-job-id             "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:callback       #'el-job-callback       "2.3.0 (2025-03-16)")
 (define-obsolete-function-alias 'el-job:n-cores-to-use #'el-job-n-cores-to-use "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:ready          #'el-job-ready "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:busy           #'el-job-busy "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:stderr         #'el-job-stderr "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:timestamps     #'el-job-timestamps "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:poll-timer     #'el-job-timer "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:finish-times   #'el-job-finish-times "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:spawn-args     #'el-job-spawn-args "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:past-elapsed   #'el-job-past-elapsed "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:queued-inputs  #'el-job-queued-inputs "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:input-sets     #'el-job-input-sets "2.3.0 (2025-03-16)")
-(define-obsolete-function-alias 'el-job:result-sets    #'el-job-result-sets "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:ready          #'el-job-ready          "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:busy           #'el-job-busy           "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:stderr         #'el-job-stderr         "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:timestamps     #'el-job-timestamps     "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:poll-timer     #'el-job-timer          "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:finish-times   #'el-job-finish-times   "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:spawn-args     #'el-job-spawn-args     "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:past-elapsed   #'el-job-past-elapsed   "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:queued-inputs  #'el-job-queued-inputs  "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:input-sets     #'el-job-input-sets     "2.3.0 (2025-03-16)")
+(define-obsolete-function-alias 'el-job:result-sets    #'el-job-result-sets    "2.3.0 (2025-03-16)")
 
 (provide 'el-job)
 
